#include <iostream>
#include <iomanip>
#include <queue>
#include <mutex>
#include <condition_variable>
#include <thread>
#include <chrono>
#include <time.h>

#include <vector>
#include <queue>
#include <utility>
#include <algorithm>

using namespace std;
using namespace std::chrono;

const int DATA_SIZE = 5000000;
const int POOL_SIZE = 16;

/**
 * Shared data provided to each thread.
 */
struct ThreadData
{
    int id;
    int *pivot_count;
    vector<int> *data;
    queue<pair<int, int>> *work_queue;
    mutex *queue_mutex;
    mutex *pivot_mutex;
    mutex *print_mutex;
    condition_variable *have_work_condition;
};

/**
 * Invoke quicksort with the given thread count.
 * @param thread_count the count of threads.
 */
bool do_sort(int thread_count);

/*
 * Check whether the data is sorted.
 * @param the data.
 * @return true if sorted, false if not.
 */
bool is_sorted(vector<int>& data);

/**
 * Quicksort running in a thread.
 * @param tdata the shared thread data.
 */
void quicksort(ThreadData tdata);

/**
 * Partition the subarray bounded by the left and right indices.
 * @param data the data being sorted.
 * @param left the left index.
 * @param right the right index.
 */
int partition(int left, int right, vector<int>& data);

/**
 * Swap two data elements.
 */
void swap(const int i, const int j, vector<int>& data);

/**
 * Main
 */
int main()
{
    int thread_count;
    bool ok = true;

    // Run with thread counts 1 through the thread POOL_SIZE.
    for (thread_count = 1; ok && (thread_count <= POOL_SIZE); thread_count++)
    {
        cout << endl << "Thread count = " << thread_count << endl;
        ok = do_sort(thread_count);
    }

    cout << endl << "Done!" << endl;
    return 0;
}

bool do_sort(int thread_count)
{
    vector<int> data;
    queue<pair<int, int>> work_queue;
    thread pool[POOL_SIZE];
    mutex queue_mutex, pivot_mutex, print_mutex;
    condition_variable have_work_condition;

    int pivot_count = 0;

    ThreadData tdata { 0, &pivot_count, &data, &work_queue,
                       &pivot_mutex, &queue_mutex, &print_mutex,
                       &have_work_condition };

    data.reserve(DATA_SIZE);
    for (int i = 0; i < DATA_SIZE; i++) data.push_back(i);
    random_shuffle(data.begin(), data.end());

    cout << "Before: is sorted: " << (is_sorted(data) ? "YES" : "NO") << endl;

    // Push the first work item:
    // The left and right indices of the entire data vector.
    tdata.work_queue->push(pair<int, int>(0, DATA_SIZE - 1));

    // Start the sort timer.
    auto start_time = steady_clock::now();

    // Spawn the threads.
    for (int i = 0; i < thread_count; i++)
    {
        tdata.id = i + 1;
        pool[i] = thread(quicksort, tdata);
    }

    // Join the threads.
    for (int i = 0; i < thread_count; i++)
    {
        pool[i].join();
    }

    // Calculate and print the elapsed sort time.
    auto end_time = steady_clock::now();
    long ms = duration_cast<milliseconds>(end_time - start_time).count();
    cout << "Sorting time = " << ms/1000.0 << " seconds." << endl;
    bool ok = is_sorted(data);
    cout << "After: is sorted: " << (ok ? "YES" : "NO") << endl;

    return ok;
}

bool is_sorted(vector<int>& data)
{
    for (int i = 1; i < data.size(); i++)
    {
        if (data[i] < data[i-1]) return false;
    }

    return true;
}

void quicksort(ThreadData tdata)
{
    int& pivot_count = *tdata.pivot_count;
    vector<int>& data = *tdata.data;
    queue<pair<int, int>> *work_queue = tdata.work_queue;
    mutex& queue_mutex = *tdata.queue_mutex;
    mutex& pivot_mutex = *tdata.pivot_mutex;
    condition_variable& have_work_condition = *tdata.have_work_condition;

    this_thread::yield();

    // This thread loops until all the data is sorted.
    while (pivot_count < DATA_SIZE)
    {
        // Wait for work (a subarray to partition) to be available
        // or all the data is sorted (lambda function).
        unique_lock<mutex> qlock0(queue_mutex);
        have_work_condition.wait(qlock0, [work_queue, &pivot_count]
                                         { return !work_queue->empty()
                                               || (pivot_count == DATA_SIZE);
                                         });
        qlock0.unlock();

        // All sorted?
        bool done = false;
        unique_lock<mutex> vlock1(pivot_mutex);
        done = pivot_count == DATA_SIZE;
        vlock1.unlock();

        if (done) break;

        int left  = -1;
        int right = -1;

        // Pop off the indices of a subarray to partition.
        unique_lock<mutex> qlock1(queue_mutex);
        if (!work_queue->empty())
        {
            left  = work_queue->front().first;
            right = work_queue->front().second;
            work_queue->pop();
        }
        qlock1.unlock();

        // If the work queue was empty, go back to sleep
        // at the top of the loop.
        if (left == -1) continue;
        this_thread::yield();

        // Update the pivot count.
        unique_lock<mutex> vlock2(pivot_mutex);
        pivot_count++;
        vlock2.unlock();

        int pivot_index = partition(left, right, data);
        this_thread::yield();

        int next_left, next_right;

        // Push the indices of the left partition onto the work queue.
        next_right = pivot_index - 1;
        if ((left <= next_right) && (next_right >= 0))
        {
            pair<int, int> p1(left, next_right);
            unique_lock<mutex> qlock3(queue_mutex);
            work_queue->push(p1);
            qlock3.unlock();

            have_work_condition.notify_all();
        }

        this_thread::yield();

        // Push the indices of the right partition onto the work queue.
        next_left = pivot_index + 1;
        if ((next_left <= right) && (next_left <= right))
        {
            pair<int, int> p2(next_left, right);
            unique_lock<mutex> qlock3(queue_mutex);
            work_queue->push(p2);
            qlock3.unlock();

            have_work_condition.notify_one();
        }

        this_thread::yield();
    }

    // Each thread that finishes wakes up other sleeping threads
    // if the entire data vector is sorted.
    bool done;
    unique_lock<mutex> vlock2(pivot_mutex);
    done = pivot_count == DATA_SIZE;
    vlock2.unlock();

    if (done) have_work_condition.notify_all();
}

int partition(int left, int right, vector<int>& data)
{
    // The pivot is the leftmost element.
    // Swqp it to the rightmost position temporarily.
    int pivot = data[left];
    swap(left, right, data);

    int i = left - 1;
    int j = right;

    // Loop until the indexes cross.
    while (i < j)
    {
        // Move i to the right as long as data[i] < pivot.
        do
        {
            i++;
        } while ((i <= right) && (data[i] < pivot));

        this_thread::yield();

        // Move j to the left as long as data[j] > pivot.
        do
        {
            j--;
        } while ((j >= left) && (data[j] > pivot));

        // Swap values if necessary after both i and j stop moving.
        if (i < j) swap(i, j, data);
    }

    swap(i, right, data);  // restore the pivot
    return i;              // pivot index
}

void swap(const int i, const int j, vector<int>& data)
{
    if (i != j)
    {
        int temp = data[i];
        data[i] = data[j];
        data[j] = temp;
    }
}

bool is_sorted(vector<int>& data, int size)
{
    for (int i = 1; i < size; i++)
    {
        if (data[i] < data[i-1]) return false;
    }

    return true;
}
